/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* sampler without transposition */

#include "m_imp.h"
#include "g_canvas.h"	/* LATER fix this */

/* ------------------------- tabwrite~ -------------------------- */

static t_class *tabwrite_class;

typedef struct _tabwrite
{
    t_object x_obj;
    int x_phase;
    int x_endphase;
    float *x_vec;
    t_symbol *x_arrayname;
    t_clock *x_clock;
} t_tabwrite;

static t_int *tabwrite_perform(t_int *w)
{
    t_tabwrite *x = (t_tabwrite *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]), phase = x->x_phase, endphase = x->x_endphase;
    int nxfer = endphase - phase;
    if (!x->x_vec) goto bad;
    
    if (nxfer > n) nxfer = n;
    if (nxfer > 0)
    {
    	float *fp = x->x_vec + phase;
    	phase += nxfer;
    	while (nxfer--) *fp++ = *in++;
    	x->x_phase = phase;
    	if (phase >= endphase) clock_delay(x->x_clock, 0);
    }
bad:
    return (w+4);
}

static void tabwrite_dsp(t_tabwrite *x, t_signal **sp)
{
    int i, elemsize, yonset;
    t_field *fy;
    t_array *a = (t_array *)pd_findbyclass(x->x_arrayname, array_class);
    t_canvas *c;
    
    if (!a)
    {
    	error("%s: no such array", x->x_arrayname->s_name);
    	return;
    }
    c = (t_canvas *)array_template(a);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	|| field_type(fy) != &s_float)
    	    error("%s: needs floating-point 'y' field", x->x_arrayname->s_name);
    else if (elemsize != sizeof(float))
    	error("%s: has more than one field", x->x_arrayname->s_name);
    else
    {
    	x->x_endphase = array_npoints(a);
    	x->x_vec =  (float *)array_vec(a);
    	dsp_add(tabwrite_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
    }
}

static void tabwrite_bang(t_tabwrite *x)
{
    x->x_phase = 0;
}

static void tabwrite_tick(t_tabwrite *x)
{
    t_array *a = (t_array *)pd_findbyclass(x->x_arrayname, array_class);
    if (!a) bug("tabwrite_tick");
    else array_redraw(a);
}

static void tabwrite_free(t_tabwrite *x)
{
    clock_free(x->x_clock);
}

static void *tabwrite_new(t_symbol *s)
{
    t_tabwrite *x = (t_tabwrite *)pd_new(tabwrite_class);
    x->x_phase = 0x7fffffff;
    x->x_arrayname = s;
    x->x_clock = clock_new(x, tabwrite_tick);
    return (x);
}

static void tabwrite_setup(void)
{
    tabwrite_class = class_new(gensym("tabwrite~"), tabwrite_new, tabwrite_free,
    	sizeof(t_tabwrite), 0, A_DEFSYM, 0);
    class_addmethod(tabwrite_class, nullfn, gensym("signal"), 0);
    class_addmethod(tabwrite_class, tabwrite_dsp, gensym("dsp"), 0);
    class_addbang(tabwrite_class, tabwrite_bang);
}

/******************** TABREAD4 ***********************/

static t_class *tabread4_class;

typedef struct _tabread4
{
    t_object x_obj;
    int x_npoints;
    float *x_vec;
    t_symbol *x_arrayname;
} t_tabread4;

static t_int *tabread4_perform(t_int *w)
{
    t_tabread4 *x = (t_tabread4 *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);    
    float maxindex;
    float *buf = x->x_vec, *fp;
    int i;
    
    maxindex = x->x_npoints - 3;    /* LATER deal with truncation error? */
    if (!buf) goto zero;

#if 0
    for (i = 0,  xmax = 0, xmin = maxindex,  fp = in1; i < n; i++,  fp++)
    {
	float f = *in1;
	if (f < xmin) xmin = f;
	else if (f > xmax) xmax = f;
    }
    if (xmax < xmin + x->c_maxextent) xmax = xmin + x->c_maxextent;
    for (i = 0, splitlo = xmin+ x->c_maxextent, splithi = xmax - x->c_maxextent,
	fp = in1; i < n; i++,  fp++)
    {
	float f = *in1;
	if (f > splitlo && f < splithi) goto zero;
    }
#endif

    for (i = 0; i < n; i++)
    {
	float findex = *in++;
	int index = findex;
	float frac,  a,  b,  c,  d, cminusb;
	static int count;
	if (index < 1) index = 1, frac = 0;
	else if (index > maxindex) index = maxindex, frac = 0;
	else frac = findex - index;
	fp = buf + index;
	a = fp[-1];
	b = fp[0];
	c = fp[1];
	d = fp[2];
	/* if (!i && !(count++ & 1023))
	    post("fp = %lx,  shit = %lx,  b = %f",  fp, buf->b_shit,  b); */
	cminusb = c-b;
	*out++ = b + frac * (
	    cminusb - 0.5 * (frac-1.) * (
		(a - d + 3. * cminusb) * frac + (b - a - cminusb)
	    )
	);
    }
    return (w+5);
 zero:
    while (n--) *out++ = 0;

    return (w+5);
}

static void tabread4_dsp(t_tabread4 *x, t_signal **sp)
{
    int i, elemsize, yonset;
    t_field *fy;
    t_array *a = (t_array *)pd_findbyclass(x->x_arrayname, array_class);
    t_canvas *c;
    
    if (!a)
    {
    	error("%s: no such array", x->x_arrayname->s_name);
    	return;
    }
    c = (t_canvas *)array_template(a);
    elemsize = canvas_template_size(c);
    if (!(fy = canvas_find_field(c, gensym("y"), &yonset))
    	|| field_type(fy) != &s_float)
    	    error("%s: needs floating-point 'y' field", x->x_arrayname->s_name);
    else if (elemsize != sizeof(float))
    	error("%s: has more than one field", x->x_arrayname->s_name);
    else
    {
    	x->x_npoints = array_npoints(a);
    	x->x_vec =  (float *)array_vec(a);
    	dsp_add(tabread4_perform, 4, x,
    	    sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
    }
}

#if 0
static void tabread4_gensampname(char *buf, char *base, long int n)
{
    char b2[30];
    char *s = b2+29;
    *s = 0;
    if (n < 0) n = 0;
    while(*buf = *base++) buf++;
    if (!n) *(--s) = '0';
    else while (n)
    {
	*(--s) = '0' + (n%10);
	n /= 10;
    }
    while (*buf++ = *s++);
}

void tabread4_set(t_sigtabread4 *x, t_symbol *s)
{
    t_sigtable *wt;
    x->x_sym = s;
    if ((wt = (t_sigtable *)(s->s_thing)) &&
	ob_class(wt)->c_sym == tabread4_sigtablesym)
    {
	post("set to %s", s->s_name);
	x->x_ctl->c_buf = &wt->x_buf;
	if (wt->x_next)
	    post("tabread4(%s): warning: multiply defined", s->s_name);
    }
    else
    {
	error("tabread4: %s: no such sample", s->s_name);
	x->x_ctl->c_buf = 0;
    }
}

void tabread4_int(t_sigtabread4 *x, long int n)
{
    char buf[30];
    tabread4_gensampname(buf, "sample", n);
    tabread4_set(x, gensym(buf));
}
#endif

static void tabread4_free(t_tabread4 *x)
{
}

static void *tabread4_new(t_symbol *s)
{
    t_tabread4 *x = (t_tabread4 *)pd_new(tabread4_class);
    x->x_arrayname = s;
    outlet_new(&x->x_obj, gensym("signal"));
    return (x);
}

static void tabread4_setup(void)
{
    tabread4_class = class_new(gensym("tabread4~"), tabread4_new, tabread4_free,
    	sizeof(t_tabread4), 0, A_DEFSYM, 0);
    class_addmethod(tabread4_class, nullfn, gensym("signal"), 0);
    class_addmethod(tabread4_class, tabread4_dsp, gensym("dsp"), 0);
}

/* ------------------------ global setup routine ------------------------- */

void d_array_setup(void)
{
    tabwrite_setup();
    tabread4_setup();
}

