/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/*  miscellaneous: print~; more to come.
*/

#include "m_imp.h"  /* LATER change to m_pd.h */
#include <stdio.h>
#include <string.h>

/* ------------------------- print~ -------------------------- */
static t_class *print_class;

typedef struct _print
{
    t_object x_obj;
    t_symbol *x_sym;
    int x_count;
} t_print;

static t_int *print_perform(t_int *w)
{
    t_print *x = (t_print *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]);
    if (x->x_count)
    {
    	post("%s:", x->x_sym->s_name);
    	if (n == 1) post("%8g", in[0]);
    	else if (n == 2) post("%8g %8g", in[0], in[1]);
    	else if (n == 4) post("%8g %8g %8g %8g",
    	    in[0], in[1], in[2], in[3]);
    	else while (n > 0)
    	{
    	    post("%-8.5g %-8.5g %-8.5g %-8.5g %-8.5g %-8.5g %-8.5g %-8.5g",
    	    	in[0], in[1], in[2], in[3], in[4], in[5], in[6], in[7]);
    	    n -= 8;
    	    in += 8;
    	}
    	x->x_count--;
    }
    return (w+4);
}

static void print_dsp(t_print *x, t_signal **sp)
{
    dsp_add(print_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}

static void print_int(t_print *x, t_int n)
{
    if (n < 0) n = 0;
    x->x_count = n;
}

static void print_bang(t_print *x)
{
    x->x_count = 1;
}

static void *print_new(t_symbol *s)
{
    t_print *x = (t_print *)pd_new(print_class);
    x->x_sym = (s->s_name[0]? s : gensym("print~"));
    x->x_count = 0;
    return (x);
}

static void print_setup(void)
{
    print_class = class_new(gensym("print~"), print_new, 0,
    	sizeof(t_print), 0, A_DEFSYM, 0);
    class_addmethod(print_class, nullfn, gensym("signal"), 0);
    class_addmethod(print_class, print_dsp, gensym("dsp"), 0);
    class_addbang(print_class, print_bang);
    class_addint(print_class, print_int);
}

/* ------------------------- scope~ -------------------------- */
static t_class *scope_class;

#define SCOPESIZE 256

typedef struct _scope
{
    t_object x_obj;
    t_sample x_samps[SCOPESIZE];
    int x_phase;
    int x_drawn;
    void *x_canvas;
} t_scope;

static t_int *scope_perform(t_int *w)
{
    t_scope *x = (t_scope *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]), phase = x->x_phase;
    while (n--)
    {
    	x->x_samps[phase] = *in++;
    	phase = (phase + 1) & (SCOPESIZE-1);
    }
    x->x_phase = phase;
    return (w+4);
}

static void scope_dsp(t_scope *x, t_signal **sp)
{
    dsp_add(scope_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}

static void scope_erase(t_scope *x)
{
    if (x->x_drawn) sys_vgui(".x%x.c delete gumbo\n", x->x_canvas);
}

#define X1 10.
#define X2 20.
#define YC 5.
static void scope_bang(t_scope *x)
{
    int n, phase;
    char hugebuf[10000], *s = hugebuf;
    scope_erase(x);
    sys_vgui(".x%x.c create line 10c 5c 20c 5c -tags gumbo\n", x->x_canvas);
    sprintf(s, ".x%x.c create line ", x->x_canvas);
    s += strlen(s);
    for (n = 0, phase = x->x_phase;
    	n < SCOPESIZE; phase = ((phase+1) & (SCOPESIZE-1)), n++)
    {
    	sprintf(s, "%fc %fc ", X1 + (X2 - X1) * (float)n * (1./SCOPESIZE),
    	    YC - 5 * x->x_samps[phase]);
    	s += strlen(s);
    	/* post("phase %d", phase); */
    }
    sprintf(s, "-tags gumbo\n");
    sys_gui(hugebuf);
    x->x_drawn = 1;
}

static void scope_free(t_scope *x)
{
    scope_erase(x);
}

static void *scope_new(t_symbol *s)
{
    t_scope *x = (t_scope *)pd_new(scope_class);
    x->x_phase = 0;
    x->x_drawn = 0;
    x->x_canvas = canvas_getcurrent();
    return (x);
}

static void scope_setup(void)
{
    scope_class = class_new(gensym("scope~"), scope_new, scope_free,
    	sizeof(t_scope), 0, A_DEFSYM, 0);
    class_addmethod(scope_class, nullfn, gensym("signal"), 0);
    class_addmethod(scope_class, scope_dsp, gensym("dsp"), 0);
    class_addbang(scope_class, scope_bang);
}

/* ------------------------ global setup routine ------------------------- */

void d_misc_setup(void)
{
    print_setup();
    scope_setup();
}




