/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/*  arithmetic.
*/

#include "m_pd.h"

/* -------------------------- phasor~ ------------------------------ */
static t_class *phasor_class;

typedef t_int t_phaseint;

#define RANGE 0x10000000
#define MASK (RANGE-1)

typedef struct _phasor
{
    t_object x_obj;
    t_phaseint x_phase;
    float x_conv;
} t_phasor;

static t_int *phasor_perform(t_int *w)
{
    t_phasor *x = (t_phasor *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);
    t_phaseint phase = x->x_phase;
    float conv = x->x_conv;
    float reconv = 1./RANGE;
    while (n--)
    {
    	phase += conv * *in++;
    	phase &= MASK;
    	*out++ = phase * reconv;
    }
    x->x_phase = phase;
    return (w+5);
}

static void phasor_dsp(t_phasor *x, t_signal **sp)
{
    x->x_conv = ((float)RANGE)/sp[0]->s_sr;
    dsp_add(phasor_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void *phasor_new()
{
    t_phasor *x = (t_phasor *)pd_new(phasor_class);
    x->x_phase = 0;
    outlet_new(&x->x_obj, gensym("signal"));
    return (x);
}

static void phasor_setup(void)
{
    phasor_class = class_new(gensym("phasor~"), phasor_new, 0,
    	sizeof(t_phasor), 0, A_DEFFLOAT, 0);
    class_addmethod(phasor_class, nullfn, gensym("signal"), 0);
    class_addmethod(phasor_class, phasor_dsp, gensym("dsp"), 0);
}

/* ------------------------ cos~ ----------------------------- */

#define A1 (4 * (3.14159265/2))
#define A3 (64 * (2.5 - 3.14159265))
#define A5 (1024 * ((3.14159265/2) - 1.5))

static t_class *cos_class;

typedef struct _cos
{
    t_object x_obj;
} t_cos;

static t_int *cos_perform(t_int *w)
{
    t_float *in = (t_float *)(w[1]);
    t_float *out = (t_float *)(w[2]);
    int n = (int)(w[3]);
    while (n--)
    {
    	float f = *(in++), g, g2, g3, term1, term2;
    	if (f < 0) f -= (int)(f-1);
    	else f -= (int)f;
    	if (f > 0.5)  g = f - 0.75;
    	else g = 0.25 - f;
    	g2 = g * g;
    	g3 = g * g2;
    	*out++ = g * A1 + g3 * A3 + g2 * g3 * A5;
    }
    return (w+4);
}

static void cos_dsp(t_cos *x, t_signal **sp)
{
    dsp_add(cos_perform, 3, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void *cos_new()
{
    t_cos *x = (t_cos *)pd_new(cos_class);
    outlet_new(&x->x_obj, gensym("signal"));
    return (x);
}

static void cos_setup(void)
{
    cos_class = class_new(gensym("cos~"), cos_new, 0,
    	sizeof(t_cos), 0, A_DEFFLOAT, 0);
    class_addmethod(cos_class, nullfn, gensym("signal"), 0);
    class_addmethod(cos_class, cos_dsp, gensym("dsp"), 0);
}

/* ----------------------- global setup routine ---------------- */
void d_osc_setup(void)
{
    phasor_setup();
    cos_setup();
}

