/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include <stdlib.h>
#include "m_imp.h"
#include "t_tk.h"
#include "g_canvas.h"
#include <stdio.h>
#include <string.h>

typedef struct _tick
{
    float k_point;  	/* one point to draw a big tick at */
    float k_inc;
    int k_lperb;    	/* little ticks per big; 0 if no ticks to draw */
} t_tick;

typedef struct _graph
{
    t_glist x_glist;
    t_tick x_xtick;
    t_tick x_ytick;
    int x_nxlabels;
    t_symbol **x_xlabel;
    float x_xlabely;
    int x_nylabels;
    t_symbol **x_ylabel;
    float x_ylabelx;
} t_graph;

#define x_gobj x_glist.gl_gobj

static void graph_redraw(t_graph *x);

static void graph_bounds(t_graph *x, double x1, double y1,
    double x2, double y2)
{
    x->x_glist.gl_x1 = x1;
    x->x_glist.gl_x2 = x2;
    x->x_glist.gl_y1 = y1;
    x->x_glist.gl_y2 = y2;
    if (x->x_glist.gl_x2 == x->x_glist.gl_x1 ||
    	x->x_glist.gl_y2 == x->x_glist.gl_y1)
    {
	error("graph: empty bounds rectangle");
	x1 = y1 = 0;
	x2 = y2 = 1;
    }
    graph_redraw(x);
}

void graph_xticks(t_graph *x, int n, double point, double inc)
{
    x->x_xtick.k_point = point;
    x->x_xtick.k_inc = inc;
    x->x_xtick.k_lperb = n;
    graph_redraw(x);
}

void graph_yticks(t_graph *x, int n, double point, double inc)
{
    x->x_ytick.k_point = point;
    x->x_ytick.k_inc = inc;
    x->x_ytick.k_lperb = n;
    graph_redraw(x);
}

void graph_xlabel(t_graph *x, t_symbol *s, int argc, t_atom *argv)
{
    int i;
    if (argc < 1) error("graph_xlabel: no y value given");
    else
    {
    	x->x_xlabely = atom_getfloat(argv);
    	argv++; argc--;
    	x->x_xlabel = (t_symbol **)t_resizebytes(x->x_xlabel, 
    	    x->x_nxlabels * sizeof (t_symbol *), argc * sizeof (t_symbol *));
    	x->x_nxlabels = argc;
    	for (i = 0; i < argc; i++) x->x_xlabel[i] = atom_gensym(&argv[i]);
    }
    graph_redraw(x);
}
    
void graph_ylabel(t_graph *x, t_symbol *s, int argc, t_atom *argv)
{
    int i;
    if (argc < 1) error("graph_ylabel: no x value given");
    else
    {
    	x->x_ylabelx = atom_getfloat(argv);
    	argv++; argc--;
    	x->x_ylabel = (t_symbol **)t_resizebytes(x->x_ylabel, 
    	    x->x_nylabels * sizeof (t_symbol *), argc * sizeof (t_symbol *));
    	x->x_nylabels = argc;
    	for (i = 0; i < argc; i++) x->x_ylabel[i] = atom_gensym(&argv[i]);
    }
    graph_redraw(x);
}

/****** routines to convert pixels to X or Y value and vice versa ******/

float glist_pixelstox(t_glist *x, float xpix)
{
    return (x->gl_px1 + (x->gl_px2 - x->gl_px1)  *
    	(xpix - x->gl_x1) / (x->gl_x2 - x->gl_x1));
}

float glist_pixelstoy(t_glist *x, float ypix)
{
    return (x->gl_py1 + (x->gl_py2 - x->gl_py1)  *
    	(ypix - x->gl_y1) / (x->gl_y2 - x->gl_y1));
}

float glist_xtopixels(t_glist *x, float xval)
{
    return (x->gl_px1 + (x->gl_px2 - x->gl_px1)  *
    	(xval - x->gl_x1) / (x->gl_x2 - x->gl_x1));
}

float glist_ytopixels(t_glist *x, float yval)
{
    return (x->gl_py1 + (x->gl_py2 - x->gl_py1)  *
    	(yval - x->gl_y1) / (x->gl_y2 - x->gl_y1));
}

void graph_vis(t_gobj *g, t_glist *unused_glist, t_int vis)
{
    t_graph *x = (t_graph *)g;
    t_glist *gl = (t_glist *)g;
    if (vis)
    {
    	int i;
    	float f;
    	t_gobj *g;
    	float plotx1 = x->x_glist.gl_px1;
    	float ploty1 = x->x_glist.gl_py1;
    	float plotx2 = x->x_glist.gl_px2;
    	float ploty2 = x->x_glist.gl_py2;
    	    /* draw a rectangle around the graph */
    	sys_vgui(".x%x.c create line\
    	    %f %f %f %f %f %f %f %f %f %f -tags graph%x\n",
	    glist_getcanvas(x->x_glist.gl_owner),
    	    plotx1, ploty1, plotx1, ploty2, plotx2, ploty2,
    	    plotx2, ploty1, plotx1, ploty1, x);

    	    /* draw ticks on horizontal borders.  If lperb field is
    	    zero, this is disabled. */
    	if (x->x_xtick.k_lperb)
    	{
    	    for (i = 0, f = x->x_xtick.k_point;
    		f < 0.99 * x->x_glist.gl_x2 + 0.01*x->x_glist.gl_x1; i++,
    		    f += x->x_xtick.k_inc)
    	    {
    		float tickpix = (i % x->x_xtick.k_lperb ? 2 : 4);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    glist_xtopixels(gl, f), ploty1,
	    	    glist_xtopixels(gl, f), ploty1 - tickpix, x);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    glist_xtopixels(gl, f), ploty2,
	    	    glist_xtopixels(gl, f), ploty2 + tickpix, x);
    	    }
    	    for (i = 1, f = x->x_xtick.k_point - x->x_xtick.k_inc;
    		f > 0.99 * x->x_glist.gl_x1 + 0.01*x->x_glist.gl_x2;
    		    i++, f -= x->x_xtick.k_inc)
    	    {
    		float tickpix = (i % x->x_xtick.k_lperb ? 2 : 4);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    glist_xtopixels(gl, f), ploty1,
	    	    glist_xtopixels(gl, f), ploty1 - tickpix, x);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    glist_xtopixels(gl, f), ploty2,
	    	    glist_xtopixels(gl, f), ploty2 + tickpix, x);
    	    }
    	}

    	    /* draw ticks in vertical borders*/
    	if (x->x_ytick.k_lperb)
    	{
    	    for (i = 0, f = x->x_ytick.k_point;
    		f < 0.99 * x->x_glist.gl_y2 + 0.01*x->x_glist.gl_y1;
    		    i++, f += x->x_ytick.k_inc)
    	    {
    		float tickpix = (i % x->x_ytick.k_lperb ? 2 : 4);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    plotx1, glist_ytopixels(gl, f), 
	    	    plotx1 + tickpix, glist_ytopixels(gl, f), x);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    plotx2, glist_ytopixels(gl, f), 
	    	    plotx2 - tickpix, glist_ytopixels(gl, f), x);
    	    }
    	    for (i = 1, f = x->x_ytick.k_point - x->x_ytick.k_inc;
    		f > 0.99 * x->x_glist.gl_y1 + 0.01*x->x_glist.gl_y2;
    		    i++, f -= x->x_ytick.k_inc)
    	    {
    		float tickpix = (i % x->x_ytick.k_lperb ? 2 : 4);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    plotx1, glist_ytopixels(gl, f), 
	    	    plotx1 + tickpix, glist_ytopixels(gl, f), x);
    		sys_vgui(".x%x.c create line %f %f %f %f -tags graph%x\n",
	    	    glist_getcanvas(x->x_glist.gl_owner),
	    	    plotx2, glist_ytopixels(gl, f), 
	    	    plotx2 - tickpix, glist_ytopixels(gl, f), x);
    	    }
    	}
    	    /* draw x labels */
    	for (i = 0; i < x->x_nxlabels; i++)
    	    sys_vgui(".x%x.c create text\
    	    	%f %f -text {%s} -font %s -tags graph%x\n",
    	    	glist_getcanvas(&x->x_glist),
	    	glist_xtopixels(gl, atof(x->x_xlabel[i]->s_name)),
	    	glist_ytopixels(gl, x->x_xlabely), x->x_xlabel[i]->s_name,
	    	glist_getfont(&x->x_glist), x);

    	    /* draw y labels */
    	for (i = 0; i < x->x_nylabels; i++)
    	    sys_vgui(".x%x.c create text\
    	        %f %f -text {%s} -font %s -tags graph%x\n",
    	    	glist_getcanvas(&x->x_glist),
	    	glist_xtopixels(gl, x->x_ylabelx),
	    	glist_ytopixels(gl, atof(x->x_ylabel[i]->s_name)),
	    	x->x_ylabel[i]->s_name,
	    	glist_getfont(&x->x_glist), x);

    	    /* draw contents of graph as glist */
     	for (g = x->x_glist.gl_list; g; g = g->g_next)
     	    gobj_vis(g, &x->x_glist, 1);
    }
    else
    {
    	sys_vgui(".x%x.c delete graph%x\n",
    	    glist_getcanvas(x->x_glist.gl_owner), x);
     	for (g = x->x_glist.gl_list; g; g = g->g_next)
     	    gobj_vis(g, &x->x_glist, 0);
    }
}

static void graph_redraw(t_graph *x)
{  
    if (glist_isvisible(&x->x_glist))
    {
    	graph_vis(&x->x_gobj, 0, 0); 
    	graph_vis(&x->x_gobj, 0, 1);
    }
}

t_class *graph_class;

/* --------------------------- widget behavior  ------------------- */

static void graph_getrect(t_gobj *z, t_glist *glist,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_graph *x = (t_graph *)z;
    float x1 = x->x_glist.gl_px1, x2 = x->x_glist.gl_px2, tmp;
    float y1 = x->x_glist.gl_py1, y2 = x->x_glist.gl_py2;
    if (x1 > x2) tmp = x2, x2 = x1, x1 = tmp;
    if (y1 > y2) tmp = y2, y2 = y1, y1 = tmp;
    *xp1 = x1;
    *yp1 = y1;
    *xp2 = x2;
    *yp2 = y2;
}

static void graph_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_graph *x = (t_graph *)z;
    x->x_glist.gl_px1 += dx;
    x->x_glist.gl_py1 += dy;
    x->x_glist.gl_px2 += dx;
    x->x_glist.gl_py2 += dy;
    graph_redraw(x);
}

static void graph_select(t_gobj *z, t_glist *glist, int state)
{
    sys_vgui(".x%x.c itemconfigure graph%x -fill %s\n", glist_getcanvas(glist), 
    	z, (state? "blue" : "black"));
}

static void graph_activate(t_gobj *z, t_glist *glist, int state)
{
    /* fill in later */
}

static void graph_delete(t_gobj *z, t_glist *glist)
{
    t_graph *x = (t_graph *)z;
    t_gobj *y;
    for (y = x->x_glist.gl_list; y; y = y->g_next)
    	gobj_delete(y, &x->x_glist);
    sys_vgui(".x%x.c delete graph%x\n", glist_getcanvas(glist), x);
}

void graph_save(t_gobj *z, t_binbuf *b)
{
    t_gobj *y;
    t_graph *x = (t_graph *)z;
    binbuf_addv(b, "sssffffffff;", gensym("#X"), gensym("graph"),
    	x->x_glist.gl_name,
    	x->x_glist.gl_x1, x->x_glist.gl_y1,
    	x->x_glist.gl_x2, x->x_glist.gl_y2,
    	x->x_glist.gl_px1, x->x_glist.gl_py1,
    	x->x_glist.gl_px2, x->x_glist.gl_py2);
    for (y = x->x_glist.gl_list; y; y = y->g_next) gobj_save(y, b);
    binbuf_addv(b, "ss;", gensym("#X"), gensym("pop"));
}

extern void graph_vis(t_gobj *x, t_glist *glist, int flag);

t_widgetbehavior graph_widgetbehavior =
{
    graph_getrect,
    graph_displace,
    graph_select,
    graph_activate,
    graph_delete,
    graph_vis,
    graph_save,
};


void glist_graph(t_glist *g, t_symbol *s, int argc, t_atom *argv)
{
    static int gcount = 0;
    int zz;
    int menu = 0;
    t_symbol *sym = atom_getsymarg(0, argc, argv);   
    float x1 = atom_getfloatarg(1, argc, argv);  
    float y1 = atom_getfloatarg(2, argc, argv);  
    float x2 = atom_getfloatarg(3, argc, argv);  
    float y2 = atom_getfloatarg(4, argc, argv);  
    float px1 = atom_getfloatarg(5, argc, argv);  
    float py1 = atom_getfloatarg(6, argc, argv);  
    float px2 = atom_getfloatarg(7, argc, argv);  
    float py2 = atom_getfloatarg(8, argc, argv);  
    t_graph *x = (t_graph *)pd_new(graph_class);
    char *str;
    if (!*sym->s_name)
    {
    	char buf[40];
    	sprintf(buf, "graph%d", ++gcount);
    	sym = gensym(buf);
    	menu = 1;
    }
    else if (!strncmp((str = s->s_name), "graph", 5)
    	&& (zz = atoi(str + 5)) > gcount) gcount = zz;
    if (x1 == x2 || y1 == y2) x1 = 0, x2 = 100, y1 = -1, y2 = 1;
    if (px1 == px2 || py1 == py2) px1 = 200, py1 = 320, px2 = 600, py2 = 20;
    glist_init(&x->x_glist);
    x->x_glist.gl_name = sym;
    x->x_glist.gl_x1 = x1;
    x->x_glist.gl_x2 = x2;
    x->x_glist.gl_y1 = y1;  /* no vertical flip here.  LATER rethink this */
    x->x_glist.gl_y2 = y2;
    x->x_glist.gl_px1 = px1;
    x->x_glist.gl_py1 = py1;
    x->x_glist.gl_px2 = px2;
    x->x_glist.gl_py2 = py2;
    pd_bind(&x->x_gobj.g_pd, sym);
    x->x_glist.gl_owner = g;
    x->x_xtick.k_lperb = 0;
    x->x_ytick.k_lperb = 0;
    x->x_nxlabels = 0;
    x->x_xlabel = (t_symbol **)t_getbytes(0);
    x->x_nylabels = 0;
    x->x_ylabel = (t_symbol **)t_getbytes(0);
    if (!menu) pd_pushsym(&x->x_gobj.g_pd);
    glist_add(g, &x->x_gobj);
}

    /* find the graph most recently added to this glist; return as a glist.
    Perhaps graph and glist should be the same thing???  Anyway, if none
    exists, we'll just make one. */

t_glist *glist_findgraph(t_glist *x, float x1, float x2, float y1, float y2)
{
    t_gobj *y = 0, *z;
    for (z = x->gl_list; z; z = z->g_next)
    	if (pd_class(&z->g_pd) == graph_class) y = z;
    if (y) return ((t_glist *)y);
    vmess(&x->gl_pd, gensym("graph"), "");
    return (glist_findgraph(x, 0, 0, 0, 0));
}

void graph_free(t_graph *x)
{
    t_gobj *y;
    while (y = x->x_glist.gl_list) glist_delete(&x->x_glist, y);
    pd_unbind(&x->x_gobj.g_pd, x->x_glist.gl_name);
}

extern void glist_text(t_glist *gl, t_symbol *s, int argc, t_atom *argv);
extern t_widgetbehavior graph_widgetbehavior;
extern void graph_array(t_glist *canvas);

void g_graph_setup()
{
    graph_class = class_new(gensym("graph"), 0, graph_free,
    	sizeof(t_graph), CLASS_GOBJ, 0);
    class_setwidget(graph_class, &graph_widgetbehavior);
    class_addmethod(graph_class, graph_bounds, gensym("bounds"),
    	A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(graph_class, pd_popsym, gensym("pop"), 0);
    class_addmethod(graph_class, graph_xticks, gensym("xticks"),
    	A_FLOAT, A_FLOAT, A_INT, 0);
    class_addmethod(graph_class, graph_xlabel, gensym("xlabel"), A_GIMME, 0);
    class_addmethod(graph_class, graph_yticks, gensym("yticks"),
    	A_FLOAT, A_FLOAT, A_INT, 0);
    class_addmethod(graph_class, graph_ylabel, gensym("ylabel"), A_GIMME, 0);
    class_addmethod(graph_class, glist_text, gensym("text"),
    	A_GIMME, A_NULL);
    class_addmethod(graph_class, graph_array, gensym("array"),
    	A_SYMBOL, A_INT, A_SYMBOL, A_NULL);
}
