/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "m_imp.h"
#include "g_canvas.h"
#include "t_tk.h"

#define LMARGIN 1
#define RMARGIN 9
#define TMARGIN 3
#define BMARGIN 3

static void rtext_senditup(t_rtext *x, int firsttime);

struct _rtext
{
    char *x_buf;
    int x_bufsize;
    int x_selstart;
    int x_selend;
    int x_height;
    t_text *x_text;
    t_glist *x_glist;
    t_symbol *x_sym;	    /* LATER demote this to string */
    struct _rtext *x_next;
};

t_rtext *rtext_new(t_glist *glist, t_text *who, t_rtext *next)
{
    t_rtext *x = (t_rtext *)getbytes(sizeof *x);
    char buf[50];
    x->x_height = -1;
    x->x_text = who;
    x->x_glist = glist;
    x->x_next = glist->gl_editor->e_rtext;
    x->x_sym = 0;
    binbuf_gettext(who->te_binbuf, &x->x_buf, &x->x_bufsize);
    glist->gl_editor->e_rtext = x;
    sprintf(buf, ".x%x.t%x", glist_getcanvas(x->x_glist), x);
    x->x_sym = gensym(buf);
    rtext_senditup(x, 1);
    return (x);
}

static t_rtext *rtext_entered;


    /* rename this to rtext_free... */ 

void rtext_cutoff(t_rtext *x)
{
    sys_vgui(".x%x.c delete %s\n", glist_getcanvas(x->x_glist), 
    	x->x_sym->s_name);
    if (x->x_glist->gl_editor->e_textedfor == x)
    	x->x_glist->gl_editor->e_textedfor = 0;
    if (x->x_glist->gl_editor->e_rtext == x)
    	x->x_glist->gl_editor->e_rtext = x->x_next;
    else
    {
    	t_rtext *e2;
    	for (e2 = x->x_glist->gl_editor->e_rtext; e2; e2 = e2->x_next)
    	    if (e2->x_next == x)
    	{
    	    e2->x_next = x->x_next;
    	    break;
    	}
    }
    if (rtext_entered == x) rtext_entered = 0;
    freebytes(x->x_buf, x->x_bufsize);
    freebytes(x, sizeof *x);
}

t_symbol *rtext_getsym(t_rtext *x)
{
    return (x->x_sym);
}

void rtext_gettext(t_rtext *x, char **buf, int *bufsize)
{
    *buf = x->x_buf;
    *bufsize = x->x_bufsize;
}


/* LATER deal with tcl-significant characters */
/* LATER fix for arbitrary size */

static int lastone(char *s, int c, int n)
{
    char *s2 = s + n;
    while (s2 != s)
    {
    	s2--;
    	if (*s2 == c) return (n);
    	n--;
    }
    return (0);
}

#define UPBUFSIZE 10000
#define BOXWIDTH 60

static void rtext_senditup(t_rtext *x, int firsttime)
{
    float dispx, dispy;
    char tempbuf[UPBUFSIZE], *tp = tempbuf, *bp = x->x_buf;
    int outchars, inchars = x->x_bufsize;
    while (inchars > BOXWIDTH)
    {
    	int foundit = lastone(bp, '\n', BOXWIDTH), *spacep;
    	if (!foundit) foundit = lastone(bp, ' ', BOXWIDTH);
    	if (!foundit) foundit = BOXWIDTH;
   	strncpy(tp, bp, foundit-1);
    	bp += foundit;
    	tp += foundit;
    	inchars -= foundit;
    	tp[-1] = '\n';
    }
    strncpy(tp, bp, inchars);
    outchars = tp + inchars - tempbuf;
    dispx = x->x_text->te_xpos;
    dispy = x->x_text->te_ypos;
    if (firsttime)
    	sys_vgui("pdtk_text_new .x%x.c %s %f %f {%.*s} %s %s\n",
	    glist_getcanvas(x->x_glist), x->x_sym->s_name,
	    dispx + LMARGIN, dispy + TMARGIN,
	    outchars, tempbuf, glist_getfont(x->x_glist),
	    (glist_isselected(x->x_glist,
	    	&x->x_glist->gl_gobj)? "blue" : "black"));
    else
    {
    	sys_vgui("pdtk_text_set .x%x.c %s {%.*s}\n",
    	    glist_getcanvas(x->x_glist), x->x_sym->s_name, outchars, tempbuf);
    	text_drawborder(x->x_text, x->x_glist, x->x_sym,
	    rtext_width(x), rtext_height(x), 0);
    }
}

void rtext_retext(t_rtext *x)
{
    t_freebytes(x->x_buf, x->x_bufsize);
    binbuf_gettext(x->x_text->te_binbuf, &x->x_buf, &x->x_bufsize);
    rtext_senditup(x, 0);
}

/* find the rtext that goes with a text item */
t_rtext *glist_findrtext(t_glist *gl, t_text *who)
{
    t_rtext *x = gl->gl_editor->e_rtext;
    while (x && x->x_text != who) x = x->x_next;
    if (!x) bug("glist_findrtext");
    return (x);
}

int rtext_height(t_rtext *x)
{
    int nlines, n;
    char *s;
    for (s = x->x_buf, n = x->x_bufsize, nlines = 1; n--; s++)
    	if (*s == '\n') nlines++;
    return (nlines * sys_fontheight + (TMARGIN+BMARGIN));
}

int rtext_width(t_rtext *x)
{
    int length = 0, longest = 0, n;
    char *s;
    for (s = x->x_buf, n = x->x_bufsize; n--; s++)
    {
    	if (*s == '\n' || !n)
    	{
    	    if (length > longest) longest = length;
    	    length = 0;
    	}
    	else length++;
    }
    if (longest < 2) longest = 2;
    return (longest * sys_fontwidth + (LMARGIN+RMARGIN));
}

void rtext_displace(t_rtext *x, int dx, int dy)
{
    sys_vgui(".x%x.c move %s %d %d\n", glist_getcanvas(x->x_glist), 
    	x->x_sym->s_name, dx, dy);
}

void rtext_select(t_rtext *x, int state)
{
    t_glist *glist = x->x_glist;
    t_canvas *canvas = glist_getcanvas(glist);
    sys_vgui(".x%x.c itemconfigure %s -fill %s\n", canvas, 
    	x->x_sym->s_name, (state? "blue" : "black"));
    canvas_editing = canvas;
}

void rtext_activate(t_rtext *x, int state)
{
    t_glist *glist = x->x_glist;
    t_canvas *canvas = glist_getcanvas(glist);
    if (state)
    {
    	sys_vgui(".x%x.c select from %s 0\n", canvas, 
    	    x->x_sym->s_name);
    	sys_vgui(".x%x.c select to %s end\n", canvas, 
    	    x->x_sym->s_name);
    	glist->gl_editor->e_textedfor = x;
    	glist->gl_editor->e_textdirty = 0;
    }
    else
    {
    	sys_vgui("selection clear .x%x.c\n", canvas);
    	if (glist->gl_editor->e_textedfor == x)
    	    glist->gl_editor->e_textedfor = 0;
    }
}

void rtext_key(t_rtext *x, int n)
{
    if (n == '\r') n = '\n';
    if (n == '\b')
    {
    	if (!x->x_glist->gl_editor->e_textdirty)
    	{
    	    glist_delete(x->x_glist, &x->x_text->te_g);
    	    return;
    	}
    	else if (x->x_bufsize)
    	{
    	    x->x_buf = resizebytes(x->x_buf, x->x_bufsize, x->x_bufsize-1);
    	    x->x_bufsize = x->x_bufsize - 1;
    	    rtext_senditup(x, 0);
    	}
    }
    else if (n == '\n' || isprint(n))
    {
    	if (!x->x_glist->gl_editor->e_textdirty)
    	    x->x_buf = resizebytes(x->x_buf, x->x_bufsize, 1),
    	    	x->x_bufsize = 0;
    	else x->x_buf = resizebytes(x->x_buf, x->x_bufsize, x->x_bufsize+1);
    	x->x_buf[x->x_bufsize] = n;
    	x->x_bufsize = x->x_bufsize + 1;
    	rtext_senditup(x, 0);
    }
    x->x_glist->gl_editor->e_textdirty = 1;
}
