/* Copyright (c) 1997 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

#include "m_imp.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>

void sys_announce(void);
void pd_init(void);
int sys_argparse(int argc, char **argv);
int sys_startgui(void);
int m_scheduler(void);

int sys_loademup(char *classname);
int sys_debuglevel;
t_symbol *sys_progdir;
t_symbol *sys_openfile = &s_;
int sys_fontwidth, sys_fontheight;
t_symbol *sys_defaultfont;
static char *sys_config = 0;

/* this is called from the gui process; until then we don't know the cwd. */
void glob_initfromgui(void *dummy, t_symbol *dir, t_int x1, t_int y1)
{
    sys_fontwidth = x1;
    sys_fontheight = y1;
    if (sys_openfile != &s_) glob_evalfile(0, sys_openfile, dir);
}

/* this is called from main() in s_entry.c */
int sys_main(int argc, char **argv)
{
    sys_announce();
    pd_init();    	    	    	    	/* start the message system */
    if (sys_argparse(argc, argv)) return (1);	/* parse cmd line */
    if (sys_startgui()) return(1);  	    	/* turn on the gui */
    if (sys_init_sysdep(sys_config)) 	    	/* open audio and MIDI */
    	return (1);
    else return (m_scheduler());   	    	/* run scheduler until quits */
}

extern char glob_datestamp[];
void sys_announce(void)
{
    fprintf(stderr, "Pd version 0.18\n%s", glob_datestamp);
}

static char usagemessage[] = 
"pd: flags are:\n\
-d <debug level>\n\
-open <filename>\n";

int sys_argparse(int argc, char **argv)
{
    char sbuf[MAXPDSTRING];
    char *lastslash = strrchr(argv[0], '/');
    FILE *pipef;
    if (!lastslash) sys_progdir = gensym(".");
    else
    {
    	if (lastslash > argv[0] + (MAXPDSTRING-1))
    	    lastslash = argv[0] + (MAXPDSTRING-1);
    	strncpy(sbuf, argv[0], lastslash-argv[0]);
    	sbuf[MAXPDSTRING] = 0;
    	sys_progdir = gensym(sbuf);
    }
    argc--; argv++;
    while (argc && **argv == '-')
    {
    	if (!strcmp(*argv, "-d") && argc > 1 &&
    	    sscanf(argv[1], "%o", &sys_debuglevel) >= 1)
    	{
    	    argc -= 2;
    	    argv += 2;
    	}
    	else if (!strcmp(*argv, "-adc"))
    	{
    	    sys_config = "a";
    	    argc--; argv++;
    	}
    	else if (!strcmp(*argv, "-nosound"))
    	{
    	    sys_config = "n";
    	    argc--; argv++;
    	}
    	else if (!strcmp(*argv, "-open") && argc > 1)
    	{
    	    sys_openfile = gensym(argv[1]);
    	    argc -= 2;
    	    argv += 2;
    	}
        else if (!strcmp(*argv, "-lib") && argc > 1)
        {
            char *loadlib = strtok(argv[1], " :");          
            while(loadlib)
            {
                if (!sys_loademup(loadlib)) return(1);
                loadlib = strtok(NULL, " :");
            }
            argc -= 2;
            argv += 2;
        }
    	else if (!strcmp(*argv, "-font") && argc > 1)
    	{
    	    char fontbuf[80];
    	    sprintf(fontbuf, "{courier %s}",
    	    	argv[1]);
    	    sys_defaultfont = gensym(fontbuf);
    	    argc -= 2;
    	    argv += 2;
    	}
    	else
    	{
    	    fprintf(stderr, "%s", usagemessage);
    	    return (1);
    	}
    }
    if (!sys_defaultfont) sys_defaultfont =
    	gensym("{courier 8}");
    if (argc > 1) fprintf(stderr, "pd: extra args dropped\n");
    if (argc) sys_openfile = gensym(argv[0]);
    return (0);
}

int sys_getblksize(void)
{
    return (DACBLKSIZE);
}
